// USPTOProxy.cs
// Written by David J. Stein, Esq. See enclosed "License.html" for terms and conditions of use.
//
// This module contains the code for the USPTOProxy class library.

#region Using declarations

using System;
using System.Net;
using System.IO;
using System.Collections;
using System.Threading;
using System.Text.RegularExpressions;

#endregion

namespace USPTOProxy_Namespace {

	#region Basic classes, delegates, and enums

	public enum SearchStatusEnum { Created, Running, Complete, Aborted, Failed };
	public enum QueryResultsFormatEnum { ReferencesOnly, FullResults };

	public delegate void GetPatentCompleteDelegate(string strSerial, Patent patent);
	public delegate void GetPatentApplicationCompleteDelegate(string strSerial, PatentApplication patentApplication);
	public delegate void QueryPatentsCompleteDelegate(string strQuery, QueryResultsFormatEnum enumResultsFormat, ArrayList arrayResults);
	public delegate void QueryPatentApplicationsCompleteDelegate(string strQuery, QueryResultsFormatEnum enumResultsFormat, ArrayList arrayResults);

	public class Patent {
		public string strPatentNumber;
		public string strIssueDate;
		public string strTitle;
		public string strAbstract;
		public ArrayList arrayInventors = new ArrayList();
		public ArrayList arrayAssignees = new ArrayList();
		public string strApplicationNumber;
		public string strFileDate;
		public ArrayList arrayUSClasses = new ArrayList();
		public ArrayList arrayInternationalClasses = new ArrayList();
		public string strSearchField;
		public string strPrimaryExaminer;
		public string strLawFirm;
		public ArrayList arrayUSPatentReferences = new ArrayList();
		public ArrayList arrayForeignPatentReferences = new ArrayList();
		public ArrayList arrayOtherReferences = new ArrayList();
		public string strClaims;
		public string strDescription;
		public int iNumberOfFigures;
		public int iNumberOfClaims;
	}

	public class PatentApplication {
		public string strPatentApplicationNumber;
		public string strTitle;
		public string strAbstract;
		public ArrayList Inventors = new ArrayList();
		public string strFileDate;
		public ArrayList arrayUSClasses = new ArrayList();
		public ArrayList arrayInternationalClasses = new ArrayList();
		public string strLawFirm;
		public string strClaims;
		public string strDescription;
		public string strPublicationNumber;
		public int iNumberOfFigures;
		public int iNumberOfClaims;
		public bool bIssued;
	}

	public class Inventor {
		public string strName;
		public string strLocation;
		public Inventor(string strName, string strLocation) {
			this.strName = strName;
			this.strLocation = strLocation;
		}
	}

	public class Assignee {
		public string strName;
		public string strLocation;
		public Assignee(string strName, string strLocation) {
			this.strName = strName;
			this.strLocation = strLocation;
		}
	}

	public class QueryReference {
		public string strSerial;
		public string strTitle;
		public QueryReference(string strSerial, string strTitle) {
			this.strSerial = strSerial;
			this.strTitle = strTitle;
		}
	}

	#endregion

	public class USPTOProxy {

		#region Main functions

		public static Patent GetPatent(string strSerial) {
			// fetch patent from USPTO database
			string strCurrentURL = "http://patft.uspto.gov/netacgi/nph-Parser?Sect1=PTO1&Sect2=HITOFF&d=PALL&p=1&u=/netahtml/srchnum.htm&r=1&f=G&l=50&s1=" + strSerial + ".WKU.&OS=PN/" + strSerial + "&RS=PN/" + strSerial;
			Patent patent = null;
			while (strCurrentURL.Length > 0) {
				WebRequest request = WebRequest.Create(strCurrentURL);
				request.Timeout = 60000;  // by default, time out after 60 seconds
				try {
					WebResponse response = request.GetResponse();
					Stream stream = response.GetResponseStream();
					string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\r", "").Replace("\n", "");
					// construct Patent object out of raw page data
					patent = new Patent();
					patent.strPatentNumber = new Regex("<TITLE>United States Patent: (.*?)</TITLE>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					patent.strIssueDate = new Regex("</TD><TD ALIGN=\"RIGHT\" WIDTH=\"50%\"> <B>(.*?)<", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					patent.strTitle = new Regex("font size=.*?> (.*?)</font", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					while (patent.strTitle.IndexOf("  ") != -1)
						patent.strTitle = patent.strTitle.Replace("  ", " ");
					string strInventors = new Regex("Inventors:.*?<TD.*?>(.*?)</TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strInventors.Length > 0) {
						MatchCollection matches = new Regex("<B>(.*?)</B> \\((.*?)\\)", RegexOptions.Multiline).Matches(strInventors);
						foreach (Match match in matches)
							patent.arrayInventors.Add(new Inventor(match.Groups[1].Value.Replace(";", ",").Trim(), match.Groups[2].Value.Trim()));
					}
					string strAssignees = new Regex("Assignee.*?(<B>.*?)  </TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strAssignees.Length > 0) {
						MatchCollection matches = new Regex("<B>(.*?)</B> \\((.*?)\\)", RegexOptions.Multiline).Matches(strAssignees);
						foreach (Match match in matches)
							patent.arrayAssignees.Add(new Assignee(match.Groups[1].Value.Replace(";", ",").Trim(), match.Groups[2].Value.Trim()));
					}
					patent.strApplicationNumber = new Regex("Appl. No.:.*?<B> (.*?)</B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					patent.strFileDate = new Regex("Filed:.*?<B>(.*?)</B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					string strUSClasses = new Regex("Current U.S. Class:.*?(<B>.*?)  </TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Replace("<B>", "").Replace("</B>", "").Trim();
					if (strUSClasses.Length > 0) {
						MatchCollection matches = new Regex("([0-9/].*)", RegexOptions.Multiline).Matches(strUSClasses);
						foreach (Match match in matches)
							patent.arrayUSClasses.Add(match.Groups[1].Value.Trim());
					}
					string strInternationalClasses = new Regex("Intern.l Class:.*?60%\">(.*?)<", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strInternationalClasses.Length > 0) {
						MatchCollection matches = new Regex("([0-9/].*)", RegexOptions.Multiline).Matches(strInternationalClasses);
						foreach (Match match in matches)
							patent.arrayInternationalClasses.Add(match.Groups[1].Value.Trim());
					}
					patent.strSearchField = new Regex("Field of Search:.*?60%\">(.*?)<", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					patent.strPrimaryExaminer = new Regex("Primary Examiner:</I>(.*?)<", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Replace(";", ",").Trim();
					patent.strLawFirm = new Regex("Attorney, Agent or Firm:</I>(.*?)<", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					while (patent.strLawFirm.IndexOf("  ") != -1)
						patent.strLawFirm = patent.strLawFirm.Replace("  ", " ");
					string strReferences = new Regex("References Cited.*?<TABLE(.*?)</TABLE>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Replace("\r", "").Replace("\n", "");
					if (strReferences.Length > 0) {
						MatchCollection collectionUSPatentReferences = new Regex("PN%2F([0-9]*)", RegexOptions.Multiline).Matches(strReferences);
						foreach (Match match in collectionUSPatentReferences)
							patent.arrayUSPatentReferences.Add(match.Groups[1].Value.Trim());
						if (strReferences.IndexOf("Foreign Patent Documents") != -1) {
							string strForeignPatentReferences = strReferences.Substring(strReferences.IndexOf("Foreign Patent Documents"));
							MatchCollection collectionForeignPatentReferences = new Regex("<TD WIDTH=\"25%\">(.*?)</TD>.*?ALIGN=\"LEFT\">(.*?)[\\.<]", RegexOptions.Multiline).Matches(strForeignPatentReferences);
							foreach (Match match in collectionForeignPatentReferences)
								patent.arrayForeignPatentReferences.Add(match.Groups[2].Value.Trim() + " " + match.Groups[1].Value.Trim());
						}
					}
					string strOtherReferences = new Regex("<CENTER><B>Other References</B></CENTER>.*?<TD ALIGN=LEFT><br>(.*?)</TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					while (strOtherReferences.Length > 0) {
						if (strOtherReferences.IndexOf("<br>") != -1) {
							string strOtherReference = strOtherReferences.Substring(0, strOtherReferences.IndexOf("<br>"));
							while (strOtherReference.IndexOf("  ") != -1)
								strOtherReference = strOtherReference.Replace("  ", " ");
							patent.arrayOtherReferences.Add(strOtherReference.Trim());
							strOtherReferences = strOtherReferences.Substring(strOtherReferences.IndexOf("<br>") + 4);
						} else {
							string strOtherReference = strOtherReferences;
							while (strOtherReference.IndexOf("  ") != -1)
								strOtherReference = strOtherReference.Replace("  ", " ");
							patent.arrayOtherReferences.Add(strOtherReference.Trim());
							strOtherReferences = "";
						}
					}
					patent.strAbstract = new Regex("Abstract.*?<P>(.*?)</P>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					while (patent.strAbstract.IndexOf("  ") != -1)
						patent.strAbstract = patent.strAbstract.Replace("  "," ");
					patent.strAbstract = patent.strAbstract.Replace("<BR>", "\r\n").Trim();
					patent.strDescription = new Regex("<CENTER><B><I> Description.*?<HR>(.*?)<HR>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					while (patent.strDescription.IndexOf("  ") != -1)
						patent.strDescription = patent.strDescription.Replace("  "," ");
					patent.strDescription = patent.strDescription.Replace("<CENTER><B>* * * * *</B></CENTER>", "");
					patent.strDescription = patent.strDescription.Replace("<BR>", "\r\n").Trim();
					patent.strClaims = new Regex("<CENTER><B><I>Claims.*?<HR>(.*?)<HR>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					if (patent.strClaims.Length > 0)
						patent.iNumberOfClaims = new Regex("<BR><BR>(\\d)*?\\.", RegexOptions.Multiline).Matches(patent.strClaims).Count;
					else
						patent.iNumberOfClaims = 0;
					while (patent.strClaims.IndexOf("  ") != -1)
						patent.strClaims = patent.strClaims.Replace("  "," ");
					patent.strClaims = patent.strClaims.Replace("<BR>", "\r\n").Trim();
					Regex regexFigures = new Regex("FIGS?\\. (\\d*)[\\w]*? ", RegexOptions.Multiline);
					if (regexFigures.IsMatch(strPage) == true) {
						MatchCollection matches = regexFigures.Matches(strPage);
						Hashtable hashtable = new Hashtable();
						foreach (Match match in matches) {
							if (hashtable.ContainsKey(match.Groups[1].Value) == false)
								hashtable.Add(match.Groups[1].Value, "");
						}
						patent.iNumberOfFigures = hashtable.Keys.Count;
					} else
						patent.iNumberOfFigures = 0;
					strCurrentURL = "";
				} catch (Exception ex) {}
			}
			return patent;
		}

		public static PatentApplication GetPatentApplication(string strSerial) {
			// fetch patent from USPTO database
			string strCurrentURL = "http://appft1.uspto.gov/netacgi/nph-Parser?Sect1=PTO2&Sect2=HITOFF&p=1&u=%2Fnetahtml%2FPTO%2Fsearch-bool.html&r=1&f=G&l=50&co1=AND&d=PG01&s1=" + strSerial + "&OS=" + strSerial + "&RS=" + strSerial;
			PatentApplication patentApplication = null;
			while (strCurrentURL.Length > 0) {
				WebRequest request = WebRequest.Create(strCurrentURL);
				request.Timeout = 60000;  // by default, time out after 60 seconds
				try {
					WebResponse response = request.GetResponse();
					Stream stream = response.GetResponseStream();
					string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\r", "").Replace("\n", "");
					// construct Patent object out of raw page data
					patentApplication = new PatentApplication();
					patentApplication.strPatentApplicationNumber = new Regex("Serial No.*?<B> (.*?)</B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					patentApplication.strTitle = new Regex("<font size=\"\\+1\"> (.*?)</font>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					while (patentApplication.strTitle.IndexOf("  ") != -1)
						patentApplication.strTitle = patentApplication.strTitle.Replace("  ", " ");
					patentApplication.strAbstract = new Regex("Abstract.*?<P>(.*?)</P>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					while (patentApplication.strAbstract.IndexOf("  ") != -1)
						patentApplication.strAbstract = patentApplication.strAbstract.Replace("  "," ");
					patentApplication.strAbstract = patentApplication.strAbstract.Replace("<BR>", "\r\n").Trim();
					string strInventors = new Regex("Inventors:.*?<TD.*?>(.*?)</TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strInventors.Length > 0) {
						MatchCollection matches = new Regex("<B>(.*?)</B>; <I>\\((.*?)\\)</I>", RegexOptions.Multiline).Matches(strInventors);
						foreach (Match match in matches)
							patentApplication.Inventors.Add(new Inventor(match.Groups[1].Value.Replace("; ", "").Trim(), match.Groups[2].Value.Trim()));
					}
					patentApplication.strFileDate = new Regex("Filed:[\\s\\S]*?<B>(.*?)</B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					string strUSClasses = new Regex("U.S. Current Class[\\s\\S]*?<B>(.*?)</B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strUSClasses.Length > 0)
						patentApplication.arrayUSClasses.Add(strUSClasses);
					string strInternationalClasses = new Regex("Intern.l Class[\\s\\S]*?<TD[\\s\\S]*?>(.*?)</TD>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (strInternationalClasses.Length > 0)
						patentApplication.arrayInternationalClasses.Add(strInternationalClasses);
					patentApplication.strLawFirm = new Regex("Correspondence Name and Address[\\s\\S]*?<PRE>([\\s\\S]*?)</PRE>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					while (patentApplication.strLawFirm.IndexOf("  ") != -1)
						patentApplication.strLawFirm = patentApplication.strLawFirm.Replace("  ", " ");
					patentApplication.strClaims = new Regex("<CENTER><B><I>Claims[\\s\\S]*?<HR>([\\s\\S]*?)<HR>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					if (patentApplication.strClaims.Length > 0)
						patentApplication.iNumberOfClaims = new Regex("<BR><BR>(\\d)*?\\.", RegexOptions.Multiline).Matches(patentApplication.strClaims).Count;
					else
						patentApplication.iNumberOfClaims = 0;
					patentApplication.strDescription = new Regex("<CENTER><B><I>Description[\\s\\S]*?<HR>([\\s\\S]*?)<HR>", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
					while (patentApplication.strDescription.IndexOf("  ") != -1)
						patentApplication.strDescription = patentApplication.strDescription.Replace("  "," ");
					patentApplication.strDescription = patentApplication.strDescription.Replace("<CENTER><B>* * * * *</B></CENTER>", "");
					patentApplication.strDescription = patentApplication.strDescription.Replace("<BR>", "\r\n").Trim();
					patentApplication.strPublicationNumber = new Regex("<B><B><I>(\\d*?)</I></B>", RegexOptions.Multiline).Match(strPage).Groups[1].Value.Trim();
					while (patentApplication.strClaims.IndexOf("  ") != -1)
						patentApplication.strClaims = patentApplication.strClaims.Replace("  "," ");
					patentApplication.strClaims = patentApplication.strClaims.Replace("<BR>", "\r\n").Trim();
					// compute figures
					Regex regexFigures = new Regex("FIGS?\\. (\\d*)[\\w]*? ", RegexOptions.Multiline);
					if (regexFigures.IsMatch(strPage) == true) {
						MatchCollection matches = regexFigures.Matches(strPage);
						Hashtable hashtable = new Hashtable();
						foreach (Match match in matches) {
							if (hashtable.ContainsKey(match.Groups[1].Value) == false)
								hashtable.Add(match.Groups[1].Value, "");
						}
						patentApplication.iNumberOfFigures = hashtable.Keys.Count;
					} else
						patentApplication.iNumberOfFigures = 0;
					// compute issuance
					if ((patentApplication.Inventors.Count > 0) && (patentApplication.strPatentApplicationNumber.Length > 0)) {
						string strFirstInventorLastName = ((Inventor) patentApplication.Inventors[0]).strName.Substring(0, ((Inventor) patentApplication.Inventors[0]).strName.IndexOf(","));
						string queryPatentURL = @"http://patft.uspto.gov/netacgi/nph-Parser?Sect1=PTO2&Sect2=HITOFF&u=/netahtml/search-adv.htm&r=1&f=G&l=50&d=PTXT&p=1&p=1&S1=(" + patentApplication.strPatentApplicationNumber + ".APN.+AND+" + strFirstInventorLastName + ".INZZ.)&OS=apn/" + patentApplication.strPatentApplicationNumber + "+and+in/" + strFirstInventorLastName + "&RS=(APN/" + patentApplication.strPatentApplicationNumber + "+AND+IN/" + strFirstInventorLastName + ")";
						WebRequest request2 =  WebRequest.Create(queryPatentURL);
						request2.Timeout = 600000;
						try {
							Stream patentstream = request2.GetResponse().GetResponseStream();
							string strPatentResult = (new StreamReader(patentstream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\n", "\r\n");
							patentApplication.bIssued = (strPatentResult.IndexOf("United States Patent: ") == -1 ? false : true);
						} catch (Exception ex) {}
					}
					// exit
					strCurrentURL = "";
				} catch (Exception ex) {}
			}
			return patentApplication;
		}

		public static GetPatentAsync GetPatent_Start(string strSerial, GetPatentCompleteDelegate completeDelegate) {
			return new GetPatentAsync(strSerial, completeDelegate);
		}

		public static GetPatentApplicationAsync GetPatentApplication_Start(string strSerial, GetPatentApplicationCompleteDelegate completeDelegate) {
			return new GetPatentApplicationAsync(strSerial, completeDelegate);
		}

		public static QueryPatentsAsync QueryPatents_Start(string strQuery, QueryResultsFormatEnum enumResultsFormat, QueryPatentsCompleteDelegate completeDelegate) {
			return new QueryPatentsAsync(strQuery, enumResultsFormat, completeDelegate);
		}

		public static QueryPatentApplicationsAsync QueryPatentApplications_Start(string strQuery, QueryResultsFormatEnum enumResultsFormat, QueryPatentApplicationsCompleteDelegate completeDelegate) {
			return new QueryPatentApplicationsAsync(strQuery, enumResultsFormat, completeDelegate);
		}

		public static string GetPatentDatabaseFreshness() {
			string strDate = "";
			WebRequest request = WebRequest.Create("http://patft.uspto.gov/netahtml/search-adv.htm");
			request.Timeout = 60000;  // by default, time out after 60 seconds
			try {
				WebResponse response = request.GetResponse();
				Stream stream = response.GetResponseStream();
				string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\r", "").Replace("\n", "");
				strDate = new Regex("Data current through (.*?) ", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
			} catch (Exception ex) {}
			return strDate;
		}

		public static string GetPatentApplicationDatabaseFreshness() {
			string strDate = "";
			WebRequest request = WebRequest.Create("http://appft1.uspto.gov/netahtml/PTO/search-adv.html");
			request.Timeout = 60000;  // by default, time out after 60 seconds
			try {
				WebResponse response = request.GetResponse();
				Stream stream = response.GetResponseStream();
				string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\r", "").Replace("\n", "");
				strDate = new Regex("Data current through <Status_Date>(.*?)\\.", RegexOptions.Multiline).Match(strPage).Groups[1].Value;
				// This next bit is to deal with the strange fact that the Patent Database returns its date as a short string (12/1/2005), while the Application Database returns its date as a long string (December 1, 2005).
				// This will try to convert this string to a short date string, just for the sake of consistency.
				if (strDate.Length > 0) {
					try {
						strDate = DateTime.Parse(strDate).ToShortDateString();
					} catch (Exception ex) {}
				}
			} catch (Exception ex) {}
			return strDate;
		}

		#endregion

	}

	#region Search instance classes

	public class GetPatentAsync {
		public Patent patent = null;
		private string strSerial;
		private Thread searchThread;
		public SearchStatusEnum searchStatus = SearchStatusEnum.Created;
		public event GetPatentCompleteDelegate GetPatentCompleteEvent;

		internal GetPatentAsync(string strSerial, GetPatentCompleteDelegate completeDelegate) {
			this.strSerial = strSerial;
			this.searchThread = new Thread(new ThreadStart(Go));
			if (completeDelegate != null)
				GetPatentCompleteEvent += completeDelegate;
			searchThread.IsBackground = true;
			searchThread.Start();
		}

		internal void Go() {
			searchStatus = SearchStatusEnum.Running;
			patent = USPTOProxy.GetPatent(strSerial);
			searchStatus = (patent == null ? SearchStatusEnum.Failed : SearchStatusEnum.Complete);
			if (GetPatentCompleteEvent != null)
				GetPatentCompleteEvent(strSerial, patent);
		}

		public void Abort() {
			if ((searchThread != null) && (searchStatus == SearchStatusEnum.Running)) {
				searchThread.Abort();
				searchStatus = SearchStatusEnum.Aborted;
				patent = null;
			}
		}
	}

	public class GetPatentApplicationAsync {
		public PatentApplication patentApplication = null;
		private string strSerial;
		private Thread searchThread;
		public SearchStatusEnum searchStatus = SearchStatusEnum.Created;
		public event GetPatentApplicationCompleteDelegate GetPatentApplicationCompleteEvent;

		internal GetPatentApplicationAsync(string strSerial, GetPatentApplicationCompleteDelegate completeDelegate) {
			this.strSerial = strSerial;
			this.searchThread = new Thread(new ThreadStart(Go));
			if (completeDelegate != null)
				GetPatentApplicationCompleteEvent += completeDelegate;
			searchThread.IsBackground = true;
			searchThread.Start();
		}

		internal void Go() {
			searchStatus = SearchStatusEnum.Running;
			patentApplication = USPTOProxy.GetPatentApplication(strSerial);
			searchStatus = (patentApplication == null ? SearchStatusEnum.Failed : SearchStatusEnum.Complete);
			if (GetPatentApplicationCompleteEvent != null)
				GetPatentApplicationCompleteEvent(strSerial, patentApplication);
		}

		public void Abort() {
			if ((searchThread != null) && (searchStatus == SearchStatusEnum.Running)) {
				searchThread.Abort();
				searchStatus = SearchStatusEnum.Aborted;
				patentApplication = null;
			}
		}
	}

	public class QueryPatentsAsync {
		public string strQuery = "";
		public QueryResultsFormatEnum enumResultsFormat = QueryResultsFormatEnum.ReferencesOnly;
		public ArrayList arrayQueryResults;
		public SearchStatusEnum searchStatus = SearchStatusEnum.Created;
		public int iTotalResults, iFetchedResults;
		public event QueryPatentsCompleteDelegate QueryPatentsCompleteEvent;
		private Thread searchThread = null;

		internal QueryPatentsAsync(string strQuery, QueryResultsFormatEnum enumResultsFormat, QueryPatentsCompleteDelegate completeDelegate) {
			this.strQuery = strQuery;
			this.enumResultsFormat = enumResultsFormat;
			if (completeDelegate != null)
				QueryPatentsCompleteEvent += completeDelegate;
			arrayQueryResults = new ArrayList();
			if (strQuery.Trim().Length == 0) {
				searchStatus = SearchStatusEnum.Failed;
				if (QueryPatentsCompleteEvent != null)
					QueryPatentsCompleteEvent(strQuery, enumResultsFormat, arrayQueryResults);
			} else {
				searchThread = new Thread(new ThreadStart(Go));
				searchThread.IsBackground = true;
				searchThread.Start();
			}
		}

		internal void Go() {
			searchStatus = SearchStatusEnum.Running;
			string strParsedQuery = strQuery.Replace("/", "%2F").Replace("\"", "%22").Replace(" ","+");
			string strCurrentURL = "http://patft.uspto.gov/netacgi/nph-Parser?Sect1=PTO2&Sect2=HITOFF&u=%2Fnetahtml%2Fsearch-adv.htm&r=0&f=S&l=50&d=PTXT&RS=TTL%2Fcomputerized&Refine=Refine+Search&Refine=Refine+Search&Query=" + strParsedQuery;
			int iRetries = 0;
			while (strCurrentURL.Length > 0) {
				WebRequest request = WebRequest.Create(strCurrentURL);
				request.Timeout = 60000;
				try {
					Stream stream =request.GetResponse().GetResponseStream();
					string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\n", "\r\n");
					// note: the following Regex matches twice for every patent: first for the title, second for the number.
					Regex regexPatent = new Regex(@"<TD valign=top><A  HREF=[\s\S]*?>([\s\S]*?)</A></TD>", RegexOptions.Multiline);
					MatchCollection collection = regexPatent.Matches(strPage);
					if (enumResultsFormat == QueryResultsFormatEnum.ReferencesOnly) {  // create a QueryReference for each result and add it to the results array
						for (int i = 0; i < collection.Count; i += 2) {
							string strSerial = collection[i].Groups[1].Value.Replace(",","").Trim();
							string strTitle = collection[i+1].Groups[1].Value.Replace("\r\n", " ").Trim();
							while (strTitle.IndexOf("  ") != -1)
								strTitle = strTitle.Replace("  ", " ");
							arrayQueryResults.Add(new QueryReference(strSerial, strTitle));
						}
					}
					else if (enumResultsFormat == QueryResultsFormatEnum.FullResults) {  // fetch each resulting patent and add it to the results array
						for (int i = 0; i < collection.Count; i++) {
							Patent patent = USPTOProxy.GetPatent(collection[i].Groups[1].Value.Replace(",","").Trim());
							if (patent != null)
								arrayQueryResults.Add(patent);
						}
					}
					// get current and total numbers of matches
					Regex regexCurrent = new Regex("<BR><I>Hits.*?through.*?<strong>(.*?)\r\n");
					Match matchCurrent = regexCurrent.Match(strPage);
					if (matchCurrent.Groups[1].Value.Length > 0)
						iFetchedResults = Int32.Parse(matchCurrent.Groups[1].Value);
					Regex regexTotal = new Regex(@"out of <strong>(.*?)</strong>", RegexOptions.Multiline);
					Match matchTotal = regexTotal.Match(strPage);
					if (matchTotal.Groups[1].Value.Length > 0)
						iTotalResults = Int32.Parse(matchTotal.Groups[1].Value);
					// get link to next page
					Regex regexNextPage = new Regex(@"HREF=(.*?)>.*?NEXT_LIST", RegexOptions.Multiline);
					strCurrentURL = regexNextPage.Match(strPage).Groups[1].Value;
					if (strCurrentURL.Length > 0)
						strCurrentURL = @"http://patft.uspto.gov" + strCurrentURL;
					iRetries = 0;
				} catch (Exception ex) {
					if (++iRetries >= 10) {  // failure
						arrayQueryResults.Clear();
						searchStatus = SearchStatusEnum.Failed;
						break;
					}
				}
			}
			if (arrayQueryResults != null)
				searchStatus = SearchStatusEnum.Complete;
			if (QueryPatentsCompleteEvent != null)
				QueryPatentsCompleteEvent(strQuery, enumResultsFormat, arrayQueryResults);
		}

		public void Abort() {
			if ((searchThread != null) && (searchStatus == SearchStatusEnum.Running)) {
				searchThread.Abort();
				searchStatus = SearchStatusEnum.Aborted;
				arrayQueryResults.Clear();
			}
		}
	}

	public class QueryPatentApplicationsAsync {
		public string strQuery = "";
		public QueryResultsFormatEnum enumResultsFormat = QueryResultsFormatEnum.ReferencesOnly;
		public ArrayList arrayQueryResults;
		public SearchStatusEnum searchStatus = SearchStatusEnum.Created;
		public int iTotalResults, iFetchedResults;
		public event QueryPatentApplicationsCompleteDelegate QueryPatentApplicationsCompleteEvent;
		private Thread searchThread = null;

		internal QueryPatentApplicationsAsync(string strQuery, QueryResultsFormatEnum enumResultsFormat, QueryPatentApplicationsCompleteDelegate completeDelegate) {
			this.strQuery = strQuery;
			this.enumResultsFormat = enumResultsFormat;
			if (completeDelegate != null)
				QueryPatentApplicationsCompleteEvent += completeDelegate;
			arrayQueryResults = new ArrayList();
			if (strQuery.Trim().Length == 0) {
				searchStatus = SearchStatusEnum.Failed;
				if (QueryPatentApplicationsCompleteEvent != null)
					QueryPatentApplicationsCompleteEvent(strQuery, enumResultsFormat, arrayQueryResults);
			} else {
				searchThread = new Thread(new ThreadStart(Go));
				searchThread.IsBackground = true;
				searchThread.Start();
			}
		}

		internal void Go() {
			searchStatus = SearchStatusEnum.Running;
			string strCurrentURL = "http://appft1.uspto.gov/netacgi/nph-Parser?Sect1=PTO2&Sect2=HITOFF&u=%2Fnetahtml%2FPTO%2Fsearch-adv.html&r=0&p=1&f=S&l=50&Query=" + strQuery + "&d=PG01";
			int iRetries = 10;
			while (strCurrentURL.Length > 0) {
				WebRequest request = WebRequest.Create(strCurrentURL);
				request.Timeout = 60000;
				try {
					Stream stream =request.GetResponse().GetResponseStream();
					string strPage = (new StreamReader(stream, System.Text.Encoding.ASCII)).ReadToEnd().Replace("\n", "\r\n");
					// note: the following Regex matches twice for every patent: first for the title, second for the number.
					Regex regexPatentApplication = new Regex(@"<TD valign=top><A  HREF=[\s\S]*?>([\s\S]*?)</A></TD>", RegexOptions.Multiline);
					MatchCollection collection = regexPatentApplication.Matches(strPage);
					if (enumResultsFormat == QueryResultsFormatEnum.ReferencesOnly) {  // create a QueryReference for each result and add it to the results array
						for (int i = 0; i < collection.Count; i += 2) {
							string strSerial = collection[i].Groups[1].Value.Replace(",","").Trim();
							string strTitle = collection[i+1].Groups[1].Value.Replace("\r\n", " ").Trim();
							while (strTitle.IndexOf("  ") != -1)
								strTitle = strTitle.Replace("  ", " ");
							arrayQueryResults.Add(new QueryReference(strSerial, strTitle));
						}
					}
					else if (enumResultsFormat == QueryResultsFormatEnum.FullResults) {  // fetch each resulting patent application and add it to the results array
						for (int i = 0; i < collection.Count; i++) {
							PatentApplication patentApplication = USPTOProxy.GetPatentApplication(collection[i].Groups[1].Value.Replace(",","").Trim());
							if (patentApplication != null)
								arrayQueryResults.Add(patentApplication);
						}
					}
					// get current and total numbers of matches
					Regex regexCurrent = new Regex("<BR><I>Hits.*?through.*?<strong>(.*?)\r\n");
					Match matchCurrent = regexCurrent.Match(strPage);
					if (matchCurrent.Groups[1].Value.Length > 0)
						iFetchedResults = Int32.Parse(matchCurrent.Groups[1].Value);
					Regex regexTotal = new Regex(@"out of <strong>(.*?)</strong>", RegexOptions.Multiline);
					Match matchTotal = regexTotal.Match(strPage);
					if (matchTotal.Groups[1].Value.Length > 0)
						iTotalResults = Int32.Parse(matchTotal.Groups[1].Value);
					// get link to next page
					Regex regexNextPage = new Regex(@"HREF=(.*?)>.*?NEXT_LIST", RegexOptions.Multiline);
					strCurrentURL = regexNextPage.Match(strPage).Groups[1].Value;
					if (strCurrentURL.Length > 0)
						strCurrentURL = @"http://appft1.uspto.gov" + strCurrentURL;
					iRetries = 0;
				} catch (Exception ex) {
					if (++iRetries >= 10) {  // failure
						arrayQueryResults.Clear();
						searchStatus = SearchStatusEnum.Failed;
						break;
					}
				}
			}
			if (arrayQueryResults != null)
				searchStatus = SearchStatusEnum.Complete;
			if (QueryPatentApplicationsCompleteEvent != null)
				QueryPatentApplicationsCompleteEvent(strQuery, enumResultsFormat, arrayQueryResults);
		}

		public void Abort() {
			if ((searchThread != null) && (searchStatus == SearchStatusEnum.Running)) {
				searchThread.Abort();
				searchStatus = SearchStatusEnum.Aborted;
				arrayQueryResults.Clear();
			}
		}
	}

	#endregion

}
